'''
Ref: uhd examples
'''
import time 
import struct 
import numpy as np 
from pathlib import Path 
from datetime import datetime, timedelta

from bioview.constants import CLOCK_TIMEOUT

def _check_pairing(r_idx, t_idx, rx_cumsum, tx_cumsum, pair_list): 
    fn = lambda x, y: (np.where(x - y < 0))[0][0]
    rx_dev = fn(r_idx, rx_cumsum)
    tx_dev = fn(t_idx, tx_cumsum)
    
    return ((rx_dev, tx_dev) in pair_list) or ((tx_dev, rx_dev) in pair_list) or rx_dev == tx_dev

def get_channel_map(n_devices: int,
                    rx_per_dev: list,
                    tx_per_dev: list,
                    balance: bool = False,
                    multi_pairs: list = None
    ): 
    ''' 
    Provide base implementations of channel mappings for the following use-cases
    [1] MIMO 
    [2] Balance Signal 
    These two modifications are on top of the multi-band pairing 
    ''' 
    rx_cumsum = np.cumsum(rx_per_dev)
    tx_cumsum = np.cumsum(tx_per_dev)

    num_rxs = rx_cumsum[-1]
    num_txs = tx_cumsum[-1]
                    
    channel_map = [['' for _ in range(num_txs)] for _ in range(num_rxs)]
    
    if balance:
        rx_enabled = [True if r%2 == 0 else False for r in range(2*n_devices)]
        tx_enabled = [True if r%2 == 0 else False for r in range(2*n_devices)]
    else:
        rx_enabled = [True for _ in range(num_rxs)]
        tx_enabled = [True for _ in range(num_txs)]
    
    rx_ctr = 1
    
    for r_idx, rx_state in enumerate(rx_enabled):
        if not rx_state:
          continue
        
        tx_ctr = 1
        for t_idx, tx_state in enumerate(tx_enabled):
          if not tx_state:
            continue
          
          if multi_pairs is None or _check_pairing(r_idx, t_idx, rx_cumsum, tx_cumsum, multi_pairs): 
            channel_map[r_idx][t_idx] = f'Tx{tx_ctr}Rx{rx_ctr}'
          
          tx_ctr += 1
        
        rx_ctr += 1
    
    return channel_map

def setup_pps(usrp, pps, num_mboards):
    """Setup the PPS source."""
    if pps == "mimo":
        if num_mboards != 2:
            print('ref = "mimo" implies 2 motherboards; ' "your system has %d boards", num_mboards)
            return False
        # make mboard 1 a slave over the MIMO Cable
        usrp.set_time_source("mimo", 1)
    else:
        usrp.set_time_source(pps)
    return True

def setup_ref(usrp, ref, num_mboards):
    """Setup the reference clock."""
    if ref == "mimo":
        if num_mboards != 2:
            print('ref = "mimo" implies 2 motherboards; ' "your system has %d boards", num_mboards)
            return False
        usrp.set_clock_source("mimo", 1)
    else:
        usrp.set_clock_source(ref)

    # Lock onto clock signals for all mboards
    if ref != "internal":
        print("Now confirming lock on clock signals...")
        end_time = datetime.now() + timedelta(milliseconds=CLOCK_TIMEOUT)
        for i in range(num_mboards):
            if ref == "mimo" and i == 0:
                continue
            is_locked = usrp.get_mboard_sensor("ref_locked", i)
            while (not is_locked) and (datetime.now() < end_time):
                time.sleep(1e-3)
                is_locked = usrp.get_mboard_sensor("ref_locked", i)
            if not is_locked:
                print("Unable to confirm clock signal locked on board %d", i)
                return False
    return True

def check_channels(usrp, rx_channels, tx_channels):
    """Check that the device has sufficient RX and TX channels available."""
    # Check that each Rx channel specified is less than the number of total number of rx channels
    # the device can support
    dev_rx_channels = usrp.get_rx_num_channels()
    if not all(map((lambda chan: chan < dev_rx_channels), rx_channels)):
        print("Invalid RX channel(s) specified.")
        return [], []
        
    # Check that each Tx channel specified is less than the number of total number of tx channels
    # the device can support
    dev_tx_channels = usrp.get_tx_num_channels()
    if not all(map((lambda chan: chan < dev_tx_channels), tx_channels)):
        print("Invalid TX channel(s) specified.")
        return [], []
    
    return rx_channels, tx_channels

def read_file(fpath: str | Path, 
              channels: list,
              data_format: str = 'i8f'): 
    # Utility function to unpack stored files 
    buf_size = struct.calcsize(data_format)
    
    data_dict = {ch: [] for ch in channels}
    data_dict['Samples'] = []
     
    with open(fpath, 'rb') as f: 
        while True:
            data = f.read(buf_size)
            
            # Check for EOF
            if not data or len(data) < buf_size:
                print(f'Finished reading data')
                break
            
            # Unpack the data
            values = struct.unpack(data_format, data)
            
            # Store in a proper format
            data_dict['Samples'].append(values[0])
            for idx, ch in enumerate(channels): 
                data_dict[ch].append(complex(values[idx*2+1], values[idx*2+2])) 
    
    return data_dict


import queue
import numpy as np

from PyQt6.QtCore import QThread, pyqtSignal

from bioview.utils import init_save_file, update_save_file, get_filter, apply_filter
from bioview.constants import SAVE_BUFFER_SIZE
from bioview.types import UsrpConfiguration, ExperimentConfiguration

class SaveWorker(QThread):
    data_ready = pyqtSignal(dict)
    logEvent = pyqtSignal(str, str)

    def __init__(self, 
                 exp_config: ExperimentConfiguration, 
                 usrp_config: list[UsrpConfiguration],
                 rx_queues: list[queue.Queue], 
                 disp_queue: queue.Queue, 
                 running: bool = True,
                 saving: bool = True, 
                 save_iq: bool = True,
                 buffer_size: int = 2
        ):
        super().__init__()
        self.usrp_config = usrp_config
        self.exp_config = exp_config        
        self.rx_queues = rx_queues
        self.disp_queue = disp_queue
        
        self.running = running
        self.saving = saving 
        
        # Store a few elements in buffer before adding
        self.buffer_size = buffer_size
        
        # Allow for saving either IQ or Amp/Phase (default)
        self.save_iq = save_iq
        
        # Load IF filters
        self.if_filts = [self._load_filter(freq) for freq in exp_config.channel_ifs]
        
        # Load output file
        self.out_file = exp_config.get_save_path() 
        if self.exp_config.save_phase:
            num_channels = 2 * len(exp_config.data_mapping)
        else:
            num_channels = len(exp_config.data_mapping)
        if self.saving:
            init_save_file(file_path = self.out_file, 
                           num_channels = num_channels, 
                           chunk_size=500)
        
        # Initialize states for all valid declared channel combinations
        self.phase_accumulator = {}
        self.filter_states = {}
        for ch_key in self.exp_config.data_mapping.keys():
            self.phase_accumulator[ch_key] = 0.0
            self.filter_states[ch_key] = None            

    def _load_filter(self, freq: float, order: int = 2): 
        bandwidth = self.exp_config.if_filter_bw
        low_cutoff = freq - bandwidth / 2
        high_cutoff = freq + bandwidth / 2
        
        filter = get_filter(bounds=[low_cutoff, high_cutoff], 
                            samp_rate=self.exp_config.samp_rate,
                            btype='band', order=order)
        return filter
    
    def _process_chunk(self, 
                       data, 
                       filter, 
                       if_freq, 
                       channel_key
        ):
        # Early return for empty data
        if len(data) == 0:
            return np.array([]), np.array([])
        
        # Store last sample for continuity checking
        if hasattr(self, 'last_samples') and channel_key in self.last_samples:
            # Check for significant discontinuity
            discontinuity = abs(data[0] - self.last_samples[channel_key])
            if discontinuity > 3 * np.std(data[:min(100, len(data))]):
                self.logEvent.emit('debug', f'Potential discontinuity detected in {channel_key}')
        
        # Store last sample for next buffer
        if not hasattr(self, 'last_samples'):
            self.last_samples = {}
        self.last_samples[channel_key] = data[-1]
        
        # Stateful filtering
        current_filter_state = self.filter_states.get(channel_key)
        filt_data, new_filter_state = apply_filter(data, filter, zi=current_filter_state)
        self.filter_states[channel_key] = new_filter_state
        
        # Get the current accumulated phase for this channel
        current_phase = self.phase_accumulator[channel_key]
        
        # Get phase for all samples
        phase_increment = 2 * np.pi * if_freq / self.exp_config.samp_rate
        phases = current_phase + np.arange(len(filt_data)) * phase_increment
        
        # Down-convert from IF to baseband with phase continuity
        downconversion = np.exp(-1j * phases)
        baseband_data = filt_data * downconversion 
        
        # Update phase accumulator for next buffer (mod 2π to prevent numerical drift)
        self.phase_accumulator[channel_key] = (phases[-1] + phase_increment)
        
        # Downsampling logic
        step = self.exp_config.save_ds
        end_idx = len(baseband_data) - step + 1
        num_windows = (end_idx + step - 1) // step  # Calculate the number of windows

        if num_windows <= 0:
            return np.array([]), np.array([])

        # Create indices for the start of each window
        start_indices = np.arange(0, end_idx, step)
        
        # Use advanced indexing to get all the windows
        window_indices = start_indices[:, np.newaxis] + np.arange(step)
        windows = baseband_data[window_indices]  

        if self.save_iq:
            first_comp = np.mean(np.real(windows), axis=1)
            second_comp = np.mean(np.imag(windows), axis=1)
        else:
            first_comp = np.mean(np.abs(windows), axis=1)
            second_comp = np.mean(np.angle(windows, True), axis=1)

        return first_comp, second_comp
    
    def _process(self, buffer):
        # Use numpy preallocated array for speed 
        num_channels = len(self.exp_config.data_mapping)
        save_list = np.empty((num_channels, int(buffer.shape[1] // self.exp_config.save_ds), 2))
        
        for r_idx, row in enumerate(self.exp_config.channel_mapping):
            x = buffer[r_idx, :]
            for t_idx, channel_key in enumerate(row): 
                channel_idx = self.exp_config.data_mapping[channel_key]
                # Pass the channel key for state tracking
                first_comp, second_comp = self._process_chunk(
                    data = x, 
                    filter = self.if_filts[t_idx], 
                    if_freq = self.exp_config.channel_ifs[t_idx], 
                    channel_key = channel_key
                )
                self.logEvent.emit('debug', f'Processed channel {channel_key} with index {channel_idx}')
                save_list[channel_idx, :, 0] = first_comp
                save_list[channel_idx, :, 1] = second_comp
        
        # Return all processed samples
        return save_list
    
    def run(self):
        self.logEvent.emit('debug', 'Saving started')
        
        # Preallocate empty buffer to get
        data_buf = []
        samples = [None] * len(self.rx_queues)
        
        while self.running:
            try:
                # Get from all queues
                if len(data_buf) < self.buffer_size: 
                    for idx, rx_q in enumerate(self.rx_queues): 
                        samples[idx] = rx_q.get()
                    data_buf.append(np.transpose(np.vstack(samples)))    
                else: 
                    # TODO: Correctly assign shapes
                    self.logEvent.emit('debug', f'Buffer size: {len(data_buf)} with elem shape {data_buf[0].shape}')
                    buffer_data = np.transpose(np.vstack(data_buf))
                    processed = self._process(buffer_data) 
                    self.logEvent.emit('debug', f'Processed data shape {processed.shape}')   
                    
                    # Add to display queue 
                    try: 
                        self.disp_queue.put(processed)
                    except queue.Empty: 
                        self.logEvent.emit('debug', 'Display Queue Empty')    
                    except queue.Full: 
                        self.logEvent.emit('debug', 'Display Queue Full')
                    
                    # Add to save queue as well (asynchronously save using save queue if performance is an issue)
                    
                    # Write to file, only if saving 
                    if self.saving:
                        update_save_file(self.out_file, processed)
                    
                    # Clear buffer 
                    data_buf = []
            except queue.Empty:
                self.logEvent.emit('debug', 'Rx Queue Empty')
                continue
            except queue.Full: 
                self.logEvent.emit('debug', 'Rx Queue Full')
                continue
            except Exception as e:
                self.logEvent.emit('error', f'Saving error: {e}')
                continue
                
        self.logEvent.emit('debug', 'Saving stopped')
        
    def stop(self):
        self.running = False


from PyQt6.QtWidgets import QGridLayout, QLabel, QGroupBox, QDoubleSpinBox
from PyQt6.QtCore import pyqtSignal

from bioview.types import UsrpConfiguration

class UsrpDeviceConfigPanel(QGroupBox):
    modifyUSRP = pyqtSignal(str, object)
    logEvent = pyqtSignal(str, str)
    
    def __init__(self, 
                 config: UsrpConfiguration,  
                 parent = None
        ):
        super().__init__(f'{config.get_param_value('device_name')} Settings', parent)
        self.config = config 
        self.param_inputs = {}
        self.init_ui()
    
    def init_ui(self):
        layout = QGridLayout()
        
        # Create parameter inputs
        param_mappings = {
            'tx_gain': {
                'disp_str': 'TX Gain (dB)', 
                'range': (0, 70), 
                'multiplier': 1, # In case units need a multiplier during storage, such as for frequency 
                'step': 1
            },
            'rx_gain': {
                'disp_str': 'RX Gain (dB)', 
                'range': (0, 70), 
                'multiplier': 1,
                'step': 0.1
            },
            'tx_amplitude': {
                'disp_str': 'IF Amplitude', 
                'range': (0, 1), 
                'multiplier': 1,
                'step': 0.01
            },
            'if_freq': {
                'disp_str': 'IF Frequency (kHz)', 
                'range': (20, 400), 
                'multiplier': 1e3, # kHz to Hz 
                'step': 0.1
            },
            'samp_rate': {
                'disp_str': 'Sample Rate (MSps)', 
                'range': (0.1, 10), 
                'multiplier': 1e6, # MSps to sps
                'step': 0.05
            },
            'carrier_freq': {
                'disp_str': 'Carrier Freq. (MHz)', 
                'range': (30, 6000), 
                'multiplier': 1e6, # MHz to Hz
                'step': 1
            },
        }
        
        row = 0
        for param_name, val in param_mappings.items(): 
            label_text = val['disp_str']
            min_val = val['range'][0]
            max_val = val['range'][1]
            multiplier = val['multiplier']
            step = val['step']
            
            # Add text for widget 
            layout.addWidget(QLabel(label_text), row, 0)
            
            # Make as many spin boxes as channels specified  
            current_values = self.config.get_param_value(param_name)
            values = current_values if isinstance(current_values, (list, tuple)) else [current_values]
            
            input_widgets = []
                
            for col, value in enumerate(values): 
                # Make current value an integer
                display_value = value / multiplier if isinstance(value, (int, float)) else value
                
                # Make widget 
                widget = QDoubleSpinBox()
                widget.setRange(min_val, max_val)
                widget.setDecimals(2)
                widget.setSingleStep(step)
                widget.setValue(display_value)
            
                # Connect signal 
                idx = col if len(values) > 1 else None
                val = value * multiplier
                widget.valueChanged.connect(
                    lambda val, param_name=param_name, idx=idx: self.modify_usrp_config(
                        param_name=param_name, 
                        value=val,
                        idx=idx
                    )
                )

                layout.addWidget(widget, row, col+1)
                input_widgets.append(widget)
            
            self.param_inputs[param_name] = input_widgets
            row += 1
        
        self.setLayout(layout)
    
    # TODO: Change this to connect params
    def modify_usrp_config(self, param_name, value, idx=None):
        # We may have lists here, these need to be handled correctly
        if idx is not None: 
            # Get value, update idx from list and emit overall
            updated_value = self.config.get_param_value(param_name)
            updated_value[idx] = value
        else:
            updated_value = value
        
        try: 
            self.config.set_param_value(param_name, updated_value)
            self.logEvent.emit('debug', f'{self.config.device_name}: Updated {param_name} to {value} successfully')
        except Exception as e:
            self.logEvent.emit('error', f'{self.config.device_name}: Updating {param_name} failed')


import qtawesome as qta
from PyQt6.QtWidgets import (
    QGridLayout, QLabel, QGroupBox, QFileDialog, QSpinBox, QListView,
    QComboBox, QLineEdit, QPushButton, QHBoxLayout, QDoubleSpinBox
)
from PyQt6.QtCore import Qt, pyqtSignal, QEvent, QModelIndex
from PyQt6.QtGui import QStandardItemModel, QStandardItem

from bioview.types import ConnectionStatus, RunningStatus
from bioview.utils import get_qcolor

class CheckableListView(QListView):
    def __init__(self, combo_box):
        super().__init__(combo_box)
        self.combo_box = combo_box

    def mouseReleaseEvent(self, event):
        index = self.indexAt(event.pos())
        if index.isValid():
            item = self.model().itemFromIndex(index)
            self.combo_box.toggle_item(item)  # Use shared method
            self.viewport().update()
            self.combo_box.showPopup()
        else:
            super().mouseReleaseEvent(event)

class CheckableComboBox(QComboBox):
    selectionChanged = pyqtSignal(str, str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self._view = CheckableListView(self)
        self.setView(self._view)
        self.setModel(QStandardItemModel(self))
        self.view().viewport().installEventFilter(self)
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)
        self.lineEdit().setPlaceholderText("Select options...")
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        self.clearEditText()

        # Update line edit display when selection changes
        self.selectionChanged.connect(self._update_line_edit_text)

    def addItem(self, text: str, checked=False):
        item = QStandardItem(text)
        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
        item.setData(Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked,
                     Qt.ItemDataRole.CheckStateRole)
        self.model().appendRow(item)
        if checked:
            self.selectionChanged.emit('add', text)  # Ensure lineEdit updates if added initially

    def toggle_item(self, item: QStandardItem):
        if item.checkState() == Qt.CheckState.Checked:
            item.setCheckState(Qt.CheckState.Unchecked)
            self.selectionChanged.emit('remove', item.text())
        else:
            item.setCheckState(Qt.CheckState.Checked)
            self.selectionChanged.emit('add', item.text())

    def select_channel(self, text: str):
        for i in range(self.model().rowCount()):
            item = self.model().item(i)
            if item.text() == text and item.checkState() != Qt.CheckState.Checked:
                self.toggle_item(item)
                break

    def unselect_channel(self, text: str):
        for i in range(self.model().rowCount()):
            item = self.model().item(i)
            if item.text() == text and item.checkState() != Qt.CheckState.Unchecked:
                self.toggle_item(item)
                break

    def checkedItems(self):
        return [
            self.model().item(i).text()
            for i in range(self.model().rowCount())
            if self.model().item(i).checkState() == Qt.CheckState.Checked
        ]

    def _update_line_edit_text(self, *_):
        checked = self.checkedItems()
        self.lineEdit().setText(", ".join(checked) if checked else "")

    def eventFilter(self, source, event):
        if event.type() == QEvent.Type.MouseButtonPress:
            index = self.view().indexAt(event.pos())
            if not index.isValid():
                self.hidePopup()
        return super().eventFilter(source, event)
    
class ExperimentSettingsPanel(QGroupBox):    
    parameterChanged = pyqtSignal(str, object)
    logEvent = pyqtSignal(str, str)
    timeWindowChanged = pyqtSignal(int)
    gridLayoutChanged = pyqtSignal(int, int)
    addChannelRequested = pyqtSignal(str)
    removeChannelRequested = pyqtSignal(str)

    def __init__(self, 
                 config,  
                 parent=None
        ):
        super().__init__('Experiment Settings', parent)
        self.config = config
        self.param_inputs = {}
        self.init_ui()
    
    def init_ui(self):
        layout = QGridLayout()
        row = 0 
        
        # Create parameter inputs
        file_name = getattr(self.config, 'file_name')
        save_dir = getattr(self.config, 'save_dir')
        
        # File Name 
        layout.addWidget(QLabel("File Name"), row, 0)
        self.file_name = QLineEdit()
        self.file_name.setText(file_name)
        self.file_name.setFocusPolicy(Qt.FocusPolicy.ClickFocus)
        self.file_name.textChanged.connect(
            lambda value: self.update_param(
                param_name='file_name', 
                value=value
            )
        )
        layout.addWidget(self.file_name, row, 1)
        row += 1        
        
        # Folder picker
        layout.addWidget(QLabel("Save Path"), row, 0)
        
        picker_layout = QHBoxLayout()
        self.folder_path = QLineEdit()
        self.folder_path.setReadOnly(True)  # Make it read-only
        self.folder_path.setText(save_dir)
        # Update folder_path as it changes 
        self.folder_path.textChanged.connect(
            lambda value: self.update_param(
                param_name='save_dir', 
                value=value
            )
        )
        picker_layout.addWidget(self.folder_path)
        
        # Button to trigger folder selection dialog
        self.browse_button = QPushButton('  Browse')
        self.browse_button.setIcon(qta.icon('fa6s.folder', color=get_qcolor('mint')))
        self.browse_button.clicked.connect(self.openFolderDialog)
        picker_layout.addWidget(self.browse_button)
        
        layout.addLayout(picker_layout, row, 1)
        row += 1        
        
        # Downsampling ratios 
        for key, val in {'save_ds': 'Saving', 'disp_ds': 'Display'}.items(): 
            layout.addWidget(QLabel(f'{val} Downsample'), row, 0)
            widget = QDoubleSpinBox()
            widget.setRange(1, 1000)
            widget.setDecimals(0)
            widget.setSingleStep(10)
            widget.setValue(getattr(self.config, key))
            # Connect to parameter changer
            widget.textChanged.connect(
                lambda value, key=key: self.update_param(
                    param_name=key, 
                    value=value
                )
            )
            layout.addWidget(widget, row, 1)
            row += 1
        
        # Display Time Length 
        layout.addWidget(QLabel("Display Time (s)"), row, 0)
        time_layout = QHBoxLayout()
        self.time_input = QSpinBox()
        self.time_input.setRange(1, 30)
        self.time_input.setValue(10) # Set initial value
        self.time_input.valueChanged.connect(self.update_display_time)
        time_layout.addWidget(self.time_input)
        layout.addLayout(time_layout, row, 1)
        row += 1
        
        # Grid Design 
        layout.addWidget(QLabel("Plot Layout"), row, 0)
        
        grid_layout = QHBoxLayout() 
        self.rows_input = QSpinBox()
        self.rows_input.setRange(1, 4)
        self.rows_input.setValue(2)
        self.rows_input.valueChanged.connect(self.update_grid)
        self.rows_input.setEnabled(True)
        grid_layout.addWidget(self.rows_input)
        
        self.cols_input = QSpinBox()
        self.cols_input.setRange(1, 3)
        self.cols_input.setValue(2)
        self.cols_input.valueChanged.connect(self.update_grid)
        self.rows_input.setEnabled(True)
        grid_layout.addWidget(self.cols_input)
        
        layout.addLayout(grid_layout, row, 1)
        row += 1
        
        # Channel selection
        layout.addWidget(QLabel("Plot Sources"), row, 0)
        self.channel_combo = CheckableComboBox()
        for x in self.config.data_mapping.keys():
            self.channel_combo.addItem(str(x))
        self.channel_combo.selectionChanged.connect(self.update_channel) 
        
        layout.addWidget(self.channel_combo, row, 1)
        
        self.setLayout(layout)
    
    # Handle theme changes 
    def _update_icons(self): 
        self.browse_button.setIcon(qta.icon('fa6s.folder', color=get_qcolor('mint')))
        
    def event(self, event): 
        if event.type() == QEvent.Type.ApplicationPaletteChange: 
            self._update_icons()
        return super().event(event)
       
    def _add_channels_to_grid(self):
        # Init grid 
        disp_channels = getattr(self.config, 'disp_channels', [])
        
        for r in range(self.rows_input.value()): 
            for c in range(self.cols_input.value()): 
                idx = r * self.cols_input.value() + c
                if idx < len(disp_channels): 
                    self.channel_combo.select_channel(disp_channels[idx])
     
    def update_display_time(self):
        self.timeWindowChanged.emit(self.time_input.value())
    
    def update_grid(self):
        self.gridLayoutChanged.emit(self.rows_input.value(), self.cols_input.value())
    
    def update_channel(self, action, source): 
        if action == 'remove':
            self.removeChannelRequested.emit(source)
        elif action == 'add':
            self.addChannelRequested.emit(source)
        else:
            return 
    
    def openFolderDialog(self):
        folder = QFileDialog.getExistingDirectory(self, "Select Folder")
        if folder:  
            # If user didn't cancel the dialog
            self.folder_path.setText(folder)
            
    def update_param(self, param_name, value):
        self.config.set_param_value(param_name, value)
    
    def update_button_states(self, connection_status, running_status):
        if connection_status == ConnectionStatus.CONNECTED: 
            self.rows_input.setEnabled(running_status == RunningStatus.STOPPED)
            self.cols_input.setEnabled(running_status == RunningStatus.STOPPED)
            self.file_name.setEnabled(running_status == RunningStatus.STOPPED)
            self.folder_path.setEnabled(running_status == RunningStatus.STOPPED)
            self.browse_button.setEnabled(running_status == RunningStatus.STOPPED)
        else: 
            self.rows_input.setEnabled(True)
            self.cols_input.setEnabled(True)


from ctypes import byref, c_double

from PyQt6.QtCore import QThread, pyqtSignal

from bioview.types import BiopacConfiguration
from bioview.utils import wrap_result_code

class ReceiveWorker(QThread):
    logEvent = pyqtSignal(str, str)

    def __init__(self, 
                 biopac,
                 config: BiopacConfiguration,  
                 bio_queue, 
                 running:bool = True,
                 parent = None
        ):
        super().__init__(parent)
        self.config = config 
        self.biopac = biopac 
        self.bio_queue = bio_queue
        self.running = running
    
    def run(self): 
        try: 
            # Start acquisition
            wrap_result_code(self.biopac.startAcquisition())
            num_channels = len(self.config.channels)
            data_buffer = (c_double * (num_channels + 1))()  # +1 for timestamp
            
            while self.running:                     
                # Get recent sample and add to queue
                if wrap_result_code(self.mpdev.getMostRecentSample(byref(data_buffer))): 
                    sample = [data_buffer[i] for i in range(num_channels + 1)]
                    self.bio_queue.put(sample)
                
        except Exception as e: 
            self.logEvent.emit('error', e)
        
    def stop(self):
        wrap_result_code(self.mpdev.stopAcquisition())
        self.running = False 
        


from PyQt6.QtCore import QMutex

from bioview.utils import get_unique_path

''' 
We make some general assumptions, specifically - 
* Each device has two working channels 
* Each device uses the default data formats 
* Each device uses internal timing reference and clock 
* Each device sends waveforms of amplitude 1
'''
BASE_USRP_CONFIG = {
    'tx_amplitude': [1, 1], 
    'rx_channels': [0, 1], 
    'tx_channels': [0, 1], 
    'rx_subdev': 'A:A A:B', 
    'tx_subdev': 'A:A A:B', 
    'cpu_format': 'fc32', 
    'wire_format': 'sc16',
    'clock': 'internal',
    'pps': 'internal',
    'if_filter_bw': 5e3,
}

class UsrpConfiguration(): 
    def __init__(self,
                 device_name: str, 
                 if_freq: list, 
                 if_bandwidth: int, 
                 rx_gain: list,
                 tx_gain: list, 
                 samp_rate: int, 
                 carrier_freq: int,
                 **kwargs
    ):
        # Add inputs
        self.device_name = device_name
        self.if_freq = if_freq
        self.if_bandwidth = if_bandwidth
        self.rx_gain = rx_gain
        self.tx_gain = tx_gain
        self.samp_rate = samp_rate
        self.carrier_freq = carrier_freq
        
        # Add base config 
        self.tx_amplitude = kwargs.get('tx_amplitude', BASE_USRP_CONFIG['tx_amplitude'])
        self.rx_channels = kwargs.get('rx_channels', BASE_USRP_CONFIG['rx_channels'])
        self.tx_channels = kwargs.get('tx_channels', BASE_USRP_CONFIG['tx_channels'])
        self.rx_subdev = kwargs.get('rx_subdev', BASE_USRP_CONFIG['rx_subdev'])
        self.tx_subdev = kwargs.get('tx_subdev', BASE_USRP_CONFIG['tx_subdev'])
        self.cpu_format = kwargs.get('cpu_format', BASE_USRP_CONFIG['cpu_format'])
        self.wire_format = kwargs.get('wire_format', BASE_USRP_CONFIG['wire_format'])
        self.clock = kwargs.get('clock', BASE_USRP_CONFIG['clock'])
        self.pps = kwargs.get('pps', BASE_USRP_CONFIG['pps'])
        
        # Set-up default absolute channel mapping, assuming single device.
        # This assumes that Tx/Rx are always used in pairs 
        # This must be updated if using MIMO with multiple USRPs
        self.absolute_channel_nums = self.tx_channels 
        
        # Add mutex to make updates thread safe
        self.mutex = QMutex()
            
    # Getters and setters to ensure continuous updates 
    def get_param_value(self, param): 
        self.mutex.lock()
        value = getattr(self, param)
        self.mutex.unlock()
        return value
    
    def set_param_value(self, param, value):
        self.mutex.lock()
        # Ensure we match types 
        current_type = type(getattr(self, param))
        setattr(self, param, current_type(value))
        self.mutex.unlock()
    
class BiopacConfiguration(): 
    def __init__(self, 
                 device_name: str, 
                 channels: list, 
                 samp_rate: int = 1000, 
                 device_type: str = 'MP36',          
                 mpdev_path: str = None
    ):
        self.device_name = device_name
        self.channels = channels
        
        self.samp_rate = samp_rate
        
        # Essential for long-term generalizability of codebase
        self.device_type = device_type
        # By default, we load from $PATH and $ROOT but a custom path may be provided  
        self.mpdev_path = mpdev_path
    
    def get_samp_time(self):
        return 1000.0 / self.samp_rate
    
    def get_channels(self): 
        # Since the API expects 16 channels, ensure we always pad to return in the appropriate format
        return self.channels + [0] * (16 - len(self.channels))
    
# A collection of (mostly) device-agnostic configuration parameters
class ExperimentConfiguration():
    def __init__(self, 
                 save_dir: str, 
                 file_name: str, 
                 save_ds: int,
                 disp_ds: int,
                 disp_filter_spec: dict,
                 disp_channels: list = None,
                 save_phase: bool = True,
                 show_phase: bool = False,
                 loop_instructions: bool = True,
                 instruction_type: str = None, 
                 instruction_file: list[str] = [], 
                 instruction_interval: int = 5000, 
                 **kwargs
        ):
        self.save_dir = save_dir
        self.file_name = file_name
        self.save_ds = save_ds
        self.disp_ds = disp_ds
        self.disp_filter_spec = disp_filter_spec        
        self.disp_channels = disp_channels
        
        # USRP-Specific Configuration Variables
        self.if_filter_bw = kwargs.get('if_filter_bw', BASE_USRP_CONFIG['if_filter_bw'])
        self.save_phase = save_phase
        self.show_phase = show_phase
        
        # Declare mappings 
        self.channel_mapping = None 
        self.channel_ifs = {}
        self.data_mapping = {}
        self.display_sources = [] # Collection of all things to display 

        ### Common functionality for instructions 
        self.loop_instructions = loop_instructions
        self.instruction_type = instruction_type # Typically audio or text
        self.instruction_file = instruction_file
        self.instruction_interval = instruction_interval 
        
        # Add mutex to make updates thread safe
        self.mutex = QMutex()
        
    def set_param_value(self, param, value):
        self.mutex.lock()
        # Ensure we match types 
        current_type = type(getattr(self, param, None))
        if current_type is not None:
            setattr(self, param, current_type(value))
        else:
            setattr(self, param, value)
        self.mutex.unlock()
    
    def get_param(self, param, default_value = None):
        self.mutex.lock()
        try:
            value = getattr(self, param)  
        except:
            value = default_value 
        self.mutex.unlock()
        return value
        
    def get_log_path(self): 
        return get_unique_path(self.save_dir, f'{self.file_name}.log')
        
    def get_save_path(self):
        return get_unique_path(self.save_dir, f'{self.file_name}.h5') 


import numpy as np 
from scipy import signal 

'''
Filtering is done similar to our LabView implementation, 
using cascaded sections of low-order Butterworth filter. 
'''

def apply_filter(data, filter, zi=None):
    if zi is None: 
        zi = signal.sosfilt_zi(filter) * data[0] if len(data) > 0 else signal.sosfilt_zi(filter)
    
    # Apply filter with state
    filtered_data, zf = signal.sosfilt(filter, data, zi=zi)
    
    return filtered_data, zf

def get_filter(bounds: list[float], 
               samp_rate: int, 
               ftype: str = 'ellip', 
               btype: str = 'band',
               order: int = 2,
    ): 
    nyquist = samp_rate / 2
    norm_bounds = np.array(bounds) / nyquist
    
    if ftype == 'ellip':
        filt_sos = signal.ellip(order, 0.01, 50, 
                                norm_bounds, 
                                btype=btype, output='sos')
    else: 
        filt_sos = signal.butter(order, norm_bounds,
                                 btype=btype, output='sos')
    
    return filt_sos


from .biopac import load_mpdev_dll, wrap_result_code
from .caches import get_usrp_address, update_usrp_address
from .filter import get_filter, apply_filter
from .usrp import get_channel_map, setup_pps, setup_ref, check_channels
from .storage import get_unique_path, init_save_file, update_save_file
from .theme import get_color_by_idx, get_color_tuple, get_qcolor


import pygame
import time 
from pathlib import Path
from PyQt6.QtCore import QThread, QObject, pyqtSignal, QMutex, QMutexLocker

from bioview.types import ExperimentConfiguration

# Each instruction handler should be its own QObject 
class AudioPlayer(QObject):
    def __init__(self, 
                 config: ExperimentConfiguration, 
                 parent=None
    ):
        super().__init__(parent=parent)
        self.instruction_file = config.get_param('instruction_file', None)
        if self.instruction_file is None or not Path(self.instruction_file).exists(): 
            raise Exception('No valid audio file found.')
        
        self.loop_instruction = config.get_param('loop_instructions', True)
        self.mutex = QMutex()
        
        # Initialize pygame mixer
        try:
            if not pygame.mixer.get_init():
                pygame.mixer.init()
        except pygame.error as e:
            print(f"Warning: pygame mixer init issue: {e}")
    
    def pre_run(self): 
        with QMutexLocker(self.mutex):
            try:
                pygame.mixer.music.load(self.instruction_file)
            except Exception as e:
                print(f"Error loading audio file: {e}")
                
    def run(self):
        with QMutexLocker(self.mutex):
            try:    
                pygame.mixer.music.play()

                # Non-blocking check for audio completion
                while pygame.mixer.music.get_busy() and self.running:
                    self.thread().msleep(100)
                    
                return not self.loop_instruction  # True if should stop                    
            except Exception as e:
                print(f"Error playing audio: {e}")
                return True
                 
    def stop(self):
        with QMutexLocker(self.mutex):
            try:
                pygame.mixer.music.stop()
            except:
                pass

class TextInstructions(QObject): 
    textUpdate = pyqtSignal(str)
    
    def __init__(self, 
                 config: ExperimentConfiguration, 
                 parent = None
    ):
        super().__init__(parent)
        f_path = Path(config.get_param('instruction_file', ''))
        
        try: 
            self.instructions = open(f_path, 'r').read().splitlines()
        except Exception as e: 
            return 

        self.current_index = 0
        self.loop_instruction = config.get_param('loop_instructions', True)
        self.interval = config.get_param('instruction_interval', 5)  # seconds    
        
    def pre_run(self):
        pass 
    
    def run(self): 
        # Check if we've reached the end
        if self.current_index >= len(self.instructions):
            if self.loop_instruction:
                self.current_index = 0
            else:
                return
        
        # Send signal to update text
        instruction_text = self.instructions[self.current_index]
        self.current_index += 1
        self.textUpdate.emit(instruction_text)
        
        # Wait for the specified interval
        time.sleep(self.interval)
        
    def stop(self): 
        pass 
    
class InstructionsWorker(QThread):
    # Signals for communication with main thread
    logEvent = pyqtSignal(str, str)
    textUpdate = pyqtSignal(str)  # Signal to update instruction text
    showDialog = pyqtSignal()     # Signal to show the dialog
    hideDialog = pyqtSignal()     # Signal to hide the dialog
    
    def __init__(self, 
                 config: ExperimentConfiguration, 
                 running: bool = False,
                 parent=None
    ):
        super().__init__(parent)
        self.config = config
        
        self.instruction_type = config.get_param('instruction_type')
        self.instruction_handler = None 
        self.running = running 
        
        try: 
            if self.instruction_type == 'text': 
                self.instruction_handler = TextInstructions(config, parent=self)
                self.instruction_handler.textUpdate.connect(self.textUpdate)
            elif self.instruction_type == 'audio': 
                self.instruction_handler = AudioPlayer(config, parent=self)
            else:
                raise Exception(f'Invalid instruction type: {self.instruction_type}')
        except Exception as e: 
            self.logEvent.emit('error', f'Unable to initialize audio player: {e}')
        
    def run(self):
        if self.instruction_handler is None:
            self.logEvent.emit('warning', 'No instruction handler available')
            return
        
        self.running = True 
        
        # This may include file-reading, etc one time tasks before running that won't loop
        if hasattr(self.instruction_handler, 'pre_run'):
            self.instruction_handler.pre_run()
        
        if self.instruction_type == 'text': 
            self.showDialog.emit()
        
        while self.running: 
            self.instruction_handler.run()
        
        self.stop()
        
    def stop(self):
        self.running = False
        
        if self.instruction_handler is not None:
            self.instruction_handler.stop()             
            
        # Hide text dialog
        if self.instruction_type == 'text':
            self.hideDialog.emit()


BIOPAC_CONNECTION_CODES = {
    1: "MPSUCCESS",
    2: "MPDRVERR",
    3: "MPDLLBUSY",
    4: "MPINVPARA",
    5: "MPNOTCON",
    6: "MPREADY",
    7: "MPWPRETRIG",
    8: "MPWTRIG",
    9: "MPBUSY",
    10: "MPNOACTCH",
    11: "MPCOMERR",
    12: "MPINVTYPE",
    13: "MPNOTINNET",
    14: "MPSMPLDLERR",
    15: "MPMEMALLOCERR",
    16: "MPSOCKERR",
    17: "MPUNDRFLOW",
    18: "MPPRESETERR",
    19: "MPPARSEERR"
}        


import uhd

from PyQt6.QtCore import QThread, pyqtSignal

from bioview.utils import setup_ref, setup_pps, check_channels, get_usrp_address, update_usrp_address 

class Controller(QThread):
    initSucceeded = pyqtSignal(uhd.usrp.MultiUSRP, object, object)
    initFailed = pyqtSignal(str)
    logEvent = pyqtSignal(str, str)
    
    def __init__(self, config, parent=None):
        super().__init__(parent)
        self.config = config
        self.usrp = None
    
    def run(self):
        # Performs device discovery and initialization in its own separate thread.
        try:
            ser = get_usrp_address(self.config.device_name)
            if ser is not None: 
                addr = uhd.find(f'serial={ser}')[0]    
            else:
                addr = uhd.find(self.config.device_name)[0]
                ser = addr['serial']
                # Store it for future use 
                update_usrp_address(self.config.device_name, ser)
            
            self.usrp = uhd.usrp.MultiUSRP(f'serial={ser},num_recv_frames=1024')
            
            # Always select the subdevice first, the channel mapping affects the other settings
            self.usrp.set_rx_subdev_spec(uhd.usrp.SubdevSpec(self.config.rx_subdev))
            self.usrp.set_tx_subdev_spec(uhd.usrp.SubdevSpec(self.config.tx_subdev))
            
            # Set the reference clock - Return on failure
            if not setup_ref(
                self.usrp, 
                self.config.clock, 
                self.usrp.get_num_mboards()
            ):
                self.initFailed.emit('Unable to lock reference clock')
                return
            self.logEvent.emit('debug', 'Reference Locked')
            
            # Set the PPS source - Return on failure
            if not setup_pps(
                self.usrp, 
                self.config.pps, 
                self.usrp.get_num_mboards()
            ):
                self.initFailed.emit('Unable to lock timing source')
                return 
            self.logEvent.emit('debug', 'Timing Source Locked')
            
            # At this point, we can assume our device has valid and locked clock and PPS
            rx_channels, tx_channels = check_channels(
                self.usrp, 
                self.config.rx_channels, 
                self.config.tx_channels
            )
            if not rx_channels and not tx_channels:
                # If the check returned two empty channel lists, that means something went wrong
                self.initFailed.emit('Mismatch between channel configuration specified and actual channels available on device')
                return
                
            self.logEvent.emit('debug', 'Channels Validated')    
            
            samp_rate = self.config.samp_rate
            carrier_freq = self.config.carrier_freq
            
            # Setup Rx channels
            rx_gain = self.config.rx_gain
            for idx, chan in enumerate(rx_channels):
                self.usrp.set_rx_rate(samp_rate, chan)
                self.usrp.set_rx_freq(carrier_freq, chan)
                self.usrp.set_rx_gain(rx_gain[idx], chan)
                self.usrp.set_rx_antenna('RX2', chan)
            self.logEvent.emit('debug', 'Rx Channels Configured')
            
            # Setup Tx channels   v
            tx_gain = self.config.tx_gain
            for chan in tx_channels:
                self.usrp.set_tx_rate(samp_rate, chan)
                self.usrp.set_tx_freq(carrier_freq, chan)
                self.usrp.set_tx_gain(tx_gain[idx], chan)
                self.usrp.set_tx_antenna('TX1', chan)
            self.logEvent.emit('debug', 'Tx Channels Configured')                
                
            # Setup streamer objects 
            stream_args = uhd.usrp.StreamArgs(
                self.config.cpu_format, 
                self.config.get_param_value('wire_format')
            )
            
            stream_args.channels = tx_channels
            self.tx_streamer = self.usrp.get_tx_stream(stream_args)
            
            stream_args.channels = rx_channels
            self.rx_streamer = self.usrp.get_rx_stream(stream_args)
            
            self.logEvent.emit('debug', f'Connected to USRP: {addr}')
            # Emit success
            self.initSucceeded.emit(self.usrp, self.tx_streamer, self.rx_streamer)
        except Exception as e: 
            self.initFailed.emit(f'Unable to initialize device: {e}')


from enum import Enum 
from bioview.utils import get_qcolor

# Handle connection as an enum for better clarity
class ConnectionStatus(Enum): 
    DISCONNECTED = ('Not Connected', get_qcolor('red'))
    CONNECTING = ('Connecting', get_qcolor('yellow'))
    CONNECTED = ('Connected', get_qcolor('green'))

class RunningStatus(Enum): 
    NOINIT = False 
    RUNNING = True 
    STOPPED = False



import uhd 
import math 
import numpy as np 
from PyQt6.QtCore import QThread, pyqtSignal

from bioview.constants import INIT_DELAY
from bioview.types import UsrpConfiguration

class TransmitWorker(QThread):
    logEvent = pyqtSignal(str, str)
    
    def __init__(self, 
                 config: UsrpConfiguration, 
                 usrp, 
                 tx_streamer, 
                 running: bool = True, 
                 parent = None
        ):
        super().__init__(parent)
        # Modifiable params
        self.tx_gain = config.get_param_value('tx_gain').copy()
        self.tx_amplitude = config.get_param_value('tx_amplitude')
        
        # Fixed params
        self.samp_rate = config.get_param_value('samp_rate')
        self.if_freq = config.get_param_value('if_freq')
        self.tx_channels = config.get_param_value('tx_channels')
        
        self._generate_tx_waveforms()    
        
        self.config = config
        self.usrp = usrp
        self.tx_streamer = tx_streamer
        self.running = running
        self.tx_buffer_size = self.tx_streamer.get_max_num_samps() 
    
    def _generate_tx_waveforms(self):
        '''
        Generate sine waves for each Tx channel, using as minimum a buffer size as possible.
        The buffer is made larger in length to be able to read circularly without causing overflow issues 
        '''
        get_buf_size = lambda x: int(self.samp_rate * x / (math.gcd(int(self.samp_rate), int(x))**2))
        get_lcm = lambda a, b: int(a * b / math.gcd(int(a), int(b)))
        
        if len(self.if_freq) == 1: 
            self.tx_waveform_size = get_buf_size(self.if_freq[0])
        else: 
            # Return the least common multiple
            self.tx_waveform_size = get_lcm(get_buf_size(self.if_freq[0]), get_buf_size(self.if_freq[1]))
            
        len_buf = 20 * self.tx_waveform_size
        
        self.tx_waveform = np.zeros((len(self.tx_channels), len_buf), dtype=np.complex64)
        
        # Generate IQ Modulated IF signals
        for idx, _ in enumerate(self.tx_channels):
            self.tx_waveform[idx] = uhd.dsp.signals.get_continuous_tone(
                self.samp_rate,
                self.if_freq[idx],
                self.tx_amplitude[idx],
                desired_size=len_buf,
                max_size=(2 * self.samp_rate),
                waveform='sine',
            )
        
    def run(self): 
        self.logEvent.emit('debug', 'Transmission Started')
        tx_metadata = uhd.types.TXMetadata()
        tx_metadata.start_of_burst = True
        tx_metadata.end_of_burst = False
        tx_metadata.has_time_spec = True
        tx_metadata.time_spec = uhd.types.TimeSpec(self.usrp.get_time_now().get_real_secs() + INIT_DELAY)
            
        while self.running:
            # Check for updated parameters 
            curr_tx_gain = self.config.get_param_value('tx_gain')
            if curr_tx_gain != self.tx_gain: 
                for chan in self.config.tx_channels:
                    self.usrp.set_tx_gain(curr_tx_gain[chan], chan)
                self.logEvent.emit('debug', f'Tx gain updated to {curr_tx_gain}. Current {self.tx_gain}')
                self.tx_gain = curr_tx_gain
                
            try:
                # Send samples
                buffer_iter = self.tx_waveform
                num_samps = self.tx_streamer.send(buffer_iter, tx_metadata)
            except RuntimeError as ex:
                self.logEvent.emit('error', f'Runtime error in transmit: {ex}')
                continue
  
            # Continue transmission
            tx_metadata.start_of_burst = False 
            tx_metadata.has_time_spec  = False 
            
            if num_samps < self.tx_buffer_size:
                self.logEvent.emit('warning', f'Tx Sent only {num_samps} samples')
        
        # End transmission
        tx_metadata.end_of_burst = True
        self.tx_streamer.send(np.zeros_like(self.tx_waveform), tx_metadata)
        self.logEvent.emit('debug', 'Transmission Stopped')
    
    def stop(self):
        self.running = False


from datetime import datetime

import qtawesome as qta
from PyQt6.QtWidgets import QGroupBox, QHBoxLayout, QPlainTextEdit, QToolButton
from PyQt6.QtCore import Qt, pyqtSignal, QEvent

from bioview.utils import get_qcolor

class AnnotateEventPanel(QGroupBox):
    logEvent = pyqtSignal(str, str)
    
    def __init__(self, 
                 config, 
                 parent=None
        ):
        super().__init__('Mark Events', parent)    
        
        self.log_path = config.get_log_path()
        
        layout = QHBoxLayout()
        self.annotation_box = QPlainTextEdit(self)
        self.annotation_box.setReadOnly(False)
        layout.addWidget(self.annotation_box)
        
        self.make_annotation_button = QToolButton()
        self.make_annotation_button.setText('Mark Event')
        self.make_annotation_button.setIcon(qta.icon('fa6s.pen-to-square', color=get_qcolor('orange')))
        self.make_annotation_button.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.make_annotation_button.setEnabled(True)
        self.make_annotation_button.clicked.connect(self.record_annotation)
        layout.addWidget(self.make_annotation_button)
        
        self.setLayout(layout)
    
    # Handle theme changes 
    def _update_icons(self): 
        self.make_annotation_button.setIcon(qta.icon('fa6s.pen-to-square', color=get_qcolor('orange')))
        
    def event(self, event): 
        if event.type() == QEvent.Type.ApplicationPaletteChange: 
            self._update_icons()
        return super().event(event)
    
    def record_annotation(self):
        # Get current time, add to text file 
        try: 
            annotation = self.annotation_box.toPlainText()
            if not annotation.strip():  # Check if the text is empty or just whitespace
                self.logEvent.emit('debug', 'No text to append')
                
            with open(self.log_path, 'a') as f:
                f.write(f'{datetime.now().strftime('%d-%m-%Y %H:%M:%S')} - {annotation}\n')
            
            self.annotation_box.clear()
        except Exception as e:
            self.logEvent.emit('error', f'An error occurred: {e}')
        

    


import queue
import numpy as np

from PyQt6.QtCore import QThread, pyqtSignal

from bioview.utils import get_filter, apply_filter
from bioview.types import ExperimentConfiguration

class DisplayWorker(QThread):
    dataReady = pyqtSignal(np.ndarray)
    logEvent = pyqtSignal(str, str)

    def __init__(self, 
                 config: ExperimentConfiguration, 
                 disp_queue: queue.Queue, 
                 running: bool = True,
                 parent = None
        ):
        super().__init__(parent)
        self.config = config 
        self.disp_filter = get_filter(
            bounds = config.disp_filter_spec['bounds'], 
            samp_rate = config.samp_rate // (config.disp_ds * config.save_ds), 
            btype = config.disp_filter_spec['btype'],
            ftype = config.disp_filter_spec['ftype']
        )
        self.disp_queue = disp_queue
        self.running = running

    def _process(self, data):
        disp_data = [None] * len(self.config.disp_channels) 
        for idx, channel_key in enumerate(self.config.disp_channels):
            channel_idx = self.config.data_mapping[channel_key]
            
            # Channel data is a 2D array of shape (num_samples, 2) representing real and imaginary components
            if self.config.show_phase: 
                channel_data = data[channel_idx, :, 1] 
            else: 
                channel_data = data[channel_idx, :, 0]
             
            # Downsample 
            channel_data = channel_data[::self.config.disp_ds]
               
            # Filter                                  
            disp_data[idx], _ = apply_filter(channel_data, self.disp_filter)
                
        # Return all processed data 
        return np.array(disp_data)
    
    def run(self):
        self.logEvent.emit('debug', 'Display started')
        
        while self.running:
            try:
                # Get samples from queue
                samples = self.disp_queue.get()                
                # Process
                processed = self._process(samples)                 
                self.dataReady.emit(processed)
            except queue.Empty:
                self.logEvent.emit('error', 'Queue Empty')
                continue
            except Exception as e:
                self.logEvent.emit('error', f'Display error: {e}')
                continue
                
        self.logEvent.emit('debug', 'Display stopped')
        
    def stop(self):
        self.running = False


import json 
from pathlib import Path

def _get_cache_file(file_name): 
    cache_file = Path.home() / '.bioview' / file_name
    
    if not cache_file.exists():
        cache_file.parent.mkdir(parents=False, exist_ok=True)
        cache_file.touch()
    
    return cache_file
    
def get_usrp_address(device_name: str): 
    cache_file = _get_cache_file('serial_maps')
    map_dict = {}
    
    try: 
        with open(cache_file, 'r') as fobj: 
            map_dict = json.load(fobj)
    except Exception as e:
        print('Cache is empty')
        return None
        
    return map_dict[device_name]

def update_usrp_address(device_name: str,
                        device_serial: str): 
    cache_file = _get_cache_file('serial_maps')
    map_dict = {}
    
    try: 
        with open(cache_file, 'r') as fobj: 
            map_dict = json.load(fobj)
    except Exception as e:
        print('Cache does not exist currently. Creating new cache file.')
    
    map_dict[device_name] = device_serial
    
    # Update 
    try: 
        with open(cache_file, 'w') as fobj: 
            json.dump(map_dict, fobj)
    except Exception as e:
        print(f'Error updating cache: {e}')
    finally:
        print('Cache updated successfully')


from PyQt6.QtWidgets import QWidget, QHBoxLayout, QLabel
from PyQt6.QtCore import QTimer, QEvent
from PyQt6.QtGui import QPainter, QColor, QPen

from bioview.types import ConnectionStatus

class LEDIndicator(QWidget):
    ''' Indicate device status using the following codes - 
    Connected: Solid Green , 
    Connecting: Blinking Yellow, 
    Disconnected: Solid Red
    '''
    def __init__(self, 
                 state=ConnectionStatus.DISCONNECTED, 
                 size: int = 12
    ):
        super().__init__()
        self.state = state
        self.size = size
        self.blink_visible = False
        self.setFixedSize(size, size)
        
        # Timer for blinking effect
        self.blink_timer = QTimer()
        self.blink_timer.timeout.connect(self.toggle_blink)
        
        self.update_state(state)
    
    def update_state(self, state):
        self.state = state
        
        if state == ConnectionStatus.CONNECTING:
            self.blink_timer.start(100) 
        else:
            self.blink_timer.stop()
            self.blink_visible = True
        
        self.update()
    
    def toggle_blink(self):
        self.blink_visible = not self.blink_visible
        self.update()
    
    def paintEvent(self, event):
        # Draw the LED circle with appropriate color
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
    
        if self.state == ConnectionStatus.CONNECTING and not self.blink_visible: 
            color = QColor(100, 100, 0) 
        else: 
            color = self.state.value[1]
        
        painter.setBrush(color)
        painter.setPen(QPen(QColor(50, 50, 50), 1))
        
        margin = 1
        painter.drawEllipse(margin, margin, self.size - 2*margin, self.size - 2*margin)
        
class DeviceStatusWidget(QWidget):
    def __init__(self, 
                 device_name, 
                 device_state = ConnectionStatus.DISCONNECTED
    ):
        super().__init__()
        self.device_name = device_name
        self.device_state = device_state
        
        # Create horizontal layout
        layout = QHBoxLayout()
        layout.setContentsMargins(5, 2, 5, 2)
        layout.setSpacing(5)
        
        self.label = QLabel(device_name)
        self.indicator = LEDIndicator(device_state)
        
        # Add widgets to layout
        layout.addWidget(self.label)
        layout.addWidget(self.indicator)
        
        self.setLayout(layout)
    
    def update_state(self, new_state):
        self.device_state = new_state
        self.indicator.update_state(new_state)
        
class DeviceStatusPanel(QWidget):
    def __init__(self, devices: dict = {}):
        super().__init__()
        self.devices = devices
        self.device_widgets = {}
        
        # Create horizontal layout for all devices
        self.layout = QHBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(15)
        
        # Add device widgets
        for device_name, device_state in self.devices.items():
            self.add_device(device_name, device_state)
        
        self.setLayout(self.layout)
    
    # Handle theme changes 
    def _update_icons(self): 
        for device_name, device_state in self.devices.items():
            self.device_widgets[device_name].update_state(device_state)
        
    def event(self, event): 
        if event.type() == QEvent.Type.ApplicationPaletteChange: 
            self._update_icons()
        return super().event(event)
    
    def add_device(self, 
                   device_name, 
                   device_state=ConnectionStatus.DISCONNECTED
    ):
        device_widget = DeviceStatusWidget(device_name, device_state)
        self.device_widgets[device_name] = device_widget
        self.layout.addWidget(device_widget)
        self.devices[device_name] = device_state
    
    def update_device_state(self, device_name, new_state):
        if device_name in self.device_widgets:
            self.device_widgets[device_name].update_state(new_state)
            self.devices[device_name] = new_state
    
    def remove_device(self, device_name):
        if device_name in self.device_widgets:
            widget = self.device_widgets[device_name]
            self.layout.removeWidget(widget)
            widget.deleteLater()
            del self.device_widgets[device_name]
            del self.devices[device_name]


from PyQt6.QtWidgets import QDialog, QVBoxLayout, QLabel
from PyQt6.QtCore import Qt, pyqtSlot

class TextDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle('Instructions')
        self.setGeometry(100, 100, 400, 300)
        self.setWindowFlags(Qt.WindowType.Window | Qt.WindowType.WindowStaysOnTopHint)
        
        # Setup UI
        layout = QVBoxLayout()
        
        self.text_display = QLabel()
        self.text_display.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.text_display.setWordWrap(True)
        self.text_display.setStyleSheet("""
            QLabel {
                padding: 20px;
                font-size: 16px;
                border-radius: 5px;
                border: 1px solid #ccc;
            }
        """)
        
        layout.addWidget(self.text_display)
        self.setLayout(layout)
        
        # Hide by default - main thread will show when needed
        self.hide()
    
    @pyqtSlot(str)
    def update_instruction_text(self, text):
        self.text_display.setText(text)
    
    @pyqtSlot(bool)
    def toggle_ui(self, enabled): 
        if enabled:
            self.show()
            self.raise_()  # Bring to front
            self.activateWindow()
        else:
            self.hide()


import h5py 
import numpy as np 
from pathlib import Path

def get_unique_path(dirname, filename): 
    f_path = Path(dirname) / filename
    base, ext = f_path.stem, f_path.suffix
    counter = 1
    while f_path.exists():
        counter += 1
        f_path = f_path.with_name(f"{base}_{counter}{ext}")
    return f_path
    
def init_save_file(file_path, 
                   num_channels: int, 
                   chunk_size: int = 2040
                ):
    with h5py.File(file_path, 'w') as f:
        f.create_dataset(
            'data',
            shape=(num_channels, 0),
            maxshape=(num_channels, None),
            dtype='float64',
            chunks=(num_channels, chunk_size)
        )
    
def update_save_file(file_path, chunk):
    save_chunk = np.vstack([chunk[:, :, 0], chunk[:, :, 1]]) # num_channels x num_samples (all real followed by all imag)
    
    with h5py.File(file_path, 'a') as f:
        dset = f['data']
        cur_cols = dset.shape[1]
        new_cols = cur_cols + save_chunk.shape[1]
        dset.resize((save_chunk.shape[0], new_cols))
        dset[:, cur_cols:new_cols] = save_chunk


import ctypes
import os 
from pathlib import Path

from bioview.constants import BIOPAC_CONNECTION_CODES

def load_mpdev_dll(custom_loc: str = None): 
    dll = None
    try: 
        dll = ctypes.CDLL('mpdev.dll')
        return
    except FileNotFoundError: 
        print('mpdev.dll is not located in $PATH')
            
    # Check custom loc 
    if custom_loc is not None: 
        dll_locs = Path(custom_loc).glob('**/mpdev.dll')
        for loc in dll_locs: 
            dll = ctypes.CDLL(loc)
            return dll     
    
    # Check root diretory
    sys_dir = Path(os.path.abspath(os.sep)) 
    dll_locs = sys_dir.glob('Program Files*/BIOPAC*/**/x64/mpdev.dll')
    
    for loc in dll_locs:
        dll = ctypes.CDLL(loc)
        return dll 

    return None 

def wrap_result_code(result, stage = ''): 
    result_code = BIOPAC_CONNECTION_CODES.get(result, 'INVALID_CODE')
    if result_code == 'MPSUCCESS':
        return True     
    else: 
        raise Exception(f'{stage} Failure - {result_code}')


import darkdetect 
from PyQt6.QtGui import QColor

from bioview.constants import COLOR_SCHEME

def get_qcolor(name): 
    return QColor(*get_color_tuple(name))
    
def get_color_tuple(name): 
    if name not in COLOR_SCHEME.keys(): 
        print('Invalid color choice. Defaulting to blue')
        name = 'blue'
         
    if darkdetect.isDark():
        return COLOR_SCHEME[name]['dark']
    else: 
        return COLOR_SCHEME[name]['light']

def get_color_by_idx(idx=0):
    idx = idx % len(COLOR_SCHEME)
    return get_color_tuple(list(COLOR_SCHEME.keys())[idx])


from ctypes import c_int, c_double, byref

from PyQt6.QtCore import QThread, pyqtSignal

from bioview.utils import load_mpdev_dll, wrap_result_code
from bioview.constants import BIOPAC_CONNECTION_CODES

class Controller(QThread):
    initSucceeded = pyqtSignal(object)
    initFailed = pyqtSignal(str)
    logEvent = pyqtSignal(str, str)

    def __init__(self, config, parent=None):
        super().__init__(parent)  
        self.config = config
        self.biopac = None 
    
    def run(self): 
        # Load the DLL
        self.biopac = load_mpdev_dll(self.config.mpdev_path)
        if self.mpdev is None: 
            self.logEvent.emit('mpdev.dll was not found. Ensure it is available in either the system path or the custom path provided')
            return 
        
        # Connect to the device - Currently, the types values are hard-coded but they needed to be taken from the device name 
        try:
            wrap_result_code(self.biopac.connectMPDev(c_int(103), c_int(10), b'auto'), 'Initialization')
            
            # Set channels
            if sum(self.config.channels) == 0:
                raise ValueError("At least one channel must be active")
            channels_array = (c_int * 16)(*self.config.get_channels())
            wrap_result_code(self.biopac.setAcqChannels(byref(channels_array)), 'Set Channels')
        
            # Set sample rate 
            wrap_result_code(self.mpdev.setSampleRate(c_double(self.config.get_sample_time())), 'Set Sample Rate')
            
            # Emit device to main app
            self.initSucceeded.emit(self.biopac)
        except Exception as e:
            self.initFailed.emit(f'Unable to initialize device: {e}')
            
    def close(self):
        # Cleanup device handler 
        try:
            wrap_result_code(self.biopac.disconnectMPDev())
        except Exception as e: 
            self.logEvent.emit('info', 'BIOPAC connection already closed')


import qtawesome as qta
from PyQt6.QtWidgets import QGroupBox, QPushButton, QHBoxLayout, QLabel, QCheckBox
from PyQt6.QtCore import pyqtSignal, QEvent

from bioview.types import ConnectionStatus, RunningStatus
from bioview.utils import get_qcolor

class AppControlPanel(QGroupBox):
    # Define signals to emit changes to connection status
    # This is left verbose for cleaner debugging
    connectionInitiated = pyqtSignal()
    startRequested = pyqtSignal()
    stopRequested = pyqtSignal() 
    saveRequested = pyqtSignal(bool)
    instructionsEnabled = pyqtSignal(bool)
    balanceRequested = pyqtSignal()
    
    def __init__(self, parent=None):
        super().__init__('Control', parent)
        self.main_window = parent
        
        # Initialize UI
        layout = QHBoxLayout()
        padlen = 8
                         
        # Connect Button
        self.connect_button = QPushButton('   Connect')
        self.connect_button.setIcon(qta.icon('fa6s.house', color=get_qcolor('purple')))
        self.connect_button.setStyleSheet(f'padding: {padlen}px;')
        self.connect_button.clicked.connect(self.on_connect_clicked)
        layout.addWidget(self.connect_button)
        
        # Start Button
        self.start_button = QPushButton('   Start')
        self.start_button.setIcon(qta.icon('fa6s.play', color=get_qcolor('green')))
        self.start_button.setStyleSheet(f'padding: {padlen}px;')
        self.start_button.setEnabled(False)
        self.start_button.clicked.connect(self.on_start_clicked)
        layout.addWidget(self.start_button)
        
        # Saving Checkbox 
        self.save_checkbox = QCheckBox(' Save ?')
        self.save_checkbox.clicked.connect(self.on_save_toggled)
        layout.addWidget(self.save_checkbox)
        
        # Instructions Checkbox (for audio/popups, etc)
        self.instructions_checkbox = QCheckBox(' Instructions ?')
        self.instructions_checkbox.clicked.connect(self.on_instructions_toggled)
        layout.addWidget(self.instructions_checkbox)
        
        # Stop Button
        self.stop_button = QPushButton('   Stop')
        self.stop_button.setIcon(qta.icon('fa6s.stop', color=get_qcolor('red')))
        self.stop_button.setStyleSheet(f'padding: {padlen}px;')
        self.stop_button.setEnabled(False)
        self.stop_button.clicked.connect(self.on_stop_clicked)
        layout.addWidget(self.stop_button)
        
        # Balance Signal Checkbox
        self.balance_signal_button = QPushButton('   Balance Signal')
        self.balance_signal_button.setIcon(qta.icon('fa6s.rotate', color=get_qcolor('blue')))
        self.balance_signal_button.setStyleSheet(f'padding: {padlen}px;')
        self.balance_signal_button.setEnabled(False)
        self.balance_signal_button.clicked.connect(self.on_balance_clicked)
        layout.addWidget(self.balance_signal_button)
        
        # Routine selection - This will be experiment specific now
        
        layout.addStretch()
        
        self.setLayout(layout)
    
    # Handle theme changes 
    def _update_icons(self): 
        self.connect_button.setIcon(qta.icon('fa6s.house', color=get_qcolor('purple')))
        self.start_button.setIcon(qta.icon('fa6s.play', color=get_qcolor('green')))
        self.stop_button.setIcon(qta.icon('fa6s.stop', color=get_qcolor('red')))
        self.balance_signal_button.setIcon(qta.icon('fa6s.rotate', color=get_qcolor('blue')))
        
    def event(self, event): 
        if event.type() == QEvent.Type.ApplicationPaletteChange: 
            self._update_icons()
        return super().event(event)
        
    def update_button_states(self, connection_status, running_status):
        if connection_status == ConnectionStatus.CONNECTED: 
            self.start_button.setEnabled(running_status == RunningStatus.STOPPED)
            self.stop_button.setEnabled(running_status == RunningStatus.RUNNING)
            self.save_checkbox.setEnabled(running_status == RunningStatus.STOPPED)
            self.connect_button.setEnabled(False)
            self.balance_signal_button.setEnabled(True) 
        elif connection_status == ConnectionStatus.DISCONNECTED: 
            self.connect_button.setEnabled(True)
            self.start_button.setEnabled(False)
            self.save_checkbox.setEnabled(True)
            self.stop_button.setEnabled(False)
            self.balance_signal_button.setEnabled(False)
        elif connection_status == ConnectionStatus.CONNECTING: 
            self.connect_button.setEnabled(False)
            self.start_button.setEnabled(False)
            self.stop_button.setEnabled(False)
    
    def on_connect_clicked(self):
        self.connectionInitiated.emit()
    
    def on_start_clicked(self):
        self.startRequested.emit()
    
    def on_stop_clicked(self):
        self.stopRequested.emit()
    
    def on_balance_clicked(self):
        self.balanceRequested.emit()
        
    def on_save_toggled(self): 
        if self.save_checkbox.isChecked(): 
            self.saveRequested.emit(True)
        else:
            self.saveRequested.emit(False)
    
    def on_instructions_toggled(self): 
        if self.instructions_checkbox.isChecked(): 
            self.instructionsEnabled.emit(True)
        else:
            self.instructionsEnabled.emit(False)


import logging 
from PyQt6.QtWidgets import QGroupBox, QVBoxLayout, QPlainTextEdit
from PyQt6.QtCore import QObject, pyqtSignal
    
# Make a thread safe logging handler 
class QTextEditLogger(QObject, logging.Handler):
    update_log = pyqtSignal(str)

    def __init__(self, text_box):
        super().__init__()
        self.text_box = text_box
        self.update_log.connect(self.append_text)
        self.flushOnClose = False 

    def emit(self, record):
        msg = self.format(record)
        self.update_log.emit(msg)
        
    def append_text(self, text):
        cursor = self.text_box.textCursor()
        cursor.movePosition(cursor.MoveOperation.End)
        cursor.insertText(text + '\n')
        self.text_box.setTextCursor(cursor)
        # Auto-scroll to the bottom
        self.text_box.ensureCursorVisible()
    
class LogDisplayPanel(QGroupBox):
    def __init__(self, 
                 logger, 
                 parent=None
        ):
        super().__init__('Log', parent)    
        layout = QVBoxLayout()
        self.log_text_box = QPlainTextEdit(self)
        self.log_text_box.setReadOnly(True)

        layout.addWidget(self.log_text_box)
        self.setLayout(layout)
        
        # Create and add the custom logger handler
        self.logger = logger
        self.log_handler = QTextEditLogger(self.log_text_box)
        formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s',
                                      datefmt='%H:%M:%S')
        self.log_handler.setFormatter(formatter)
        self.logger.addHandler(self.log_handler)
        
    def log_message(self, level, msg):
        log_method = getattr(self.logger, level, None)
        if log_method is not None:
            log_method(msg)


# For some ass-backward reason, both uhd and time are ESSENTIAL for the app to not crash
import uhd
import time

import os 
import queue
import logging 

from PyQt6.QtWidgets import QMainWindow, QWidget, QVBoxLayout, QHBoxLayout, QStatusBar
from PyQt6.QtGui import QIcon, QGuiApplication
from PyQt6.QtCore import QMutex

from bioview.components import UsrpDeviceConfigPanel, LogDisplayPanel, ExperimentSettingsPanel, PlotGrid, AppControlPanel, AnnotateEventPanel, DeviceStatusPanel, TextDialog
from bioview.types import ConnectionStatus, RunningStatus, UsrpConfiguration, ExperimentConfiguration
from bioview.usrp import UsrpController, UsrpReceiver, UsrpTransmitter
from bioview.common import SaveWorker, DisplayWorker, InstructionsWorker
from bioview.biopac import BiopacController
from bioview.utils import get_channel_map
    
class Viewer(QMainWindow):
    def __init__(self, 
                 exp_config: ExperimentConfiguration, 
                 usrp_config: list[UsrpConfiguration] = None,
                 bio_config = None
    ):
        super().__init__()
        self.mutex = QMutex()
        
        ### Load configurations
        self.exp_config = exp_config
        self.usrp_config = usrp_config
        self.bio_config = bio_config
        
        ### Process configurations  
        self.source_counter = 0 
        self._generate_usrp_mappings()
        self._generate_biopac_mappings()
        
        ### Track state 
        self.connection_status = ConnectionStatus.DISCONNECTED
        self.running_status = RunningStatus.NOINIT
        self.saving_status = False 
        
        ### Track instruction
        if exp_config.get_param('instruction_type') is None:
            self.enable_instructions = False
        else: 
            self.enable_instructions = True 
        
        self.instruction_dialog = None     
        if exp_config.get_param('instruction_type') == 'text':
            self.instruction_dialog = TextDialog() 
        
        ### Populate list of devices 
        self.devices = {} 
        for cfg in self.usrp_config: 
            self.devices[cfg.device_name] = ConnectionStatus.DISCONNECTED
        if bio_config is not None:
            self.devices[bio_config.device_name] = ConnectionStatus.DISCONNECTED
        
        ### Set up UI
        self._init_ui()
        
        ### USRP Specific Variables
        self.usrps = [None] * len(self.usrp_config) 
        self.tx_streamers = [None] * len(self.usrp_config) 
        self.rx_streamers = [None] * len(self.usrp_config) 
        
        ### BIOPAC Specific Variables 
        self.biopac = None
        
        ### Threads
        # USRP
        self.usrp_init_thread = [None] * len(self.usrp_config) 
        self.usrp_tx_thread = [None] * len(self.usrp_config) 
        self.usrp_rx_thread = [None] * len(self.usrp_config) 
        # BIOPAC
        self.bio_init_thread = None 
        self.bio_rx_thread = None 
        # Common
        self.save_thread = None 
        self.display_thread = None 
        self.instructions_thread = None
        
        ### Data Queues
        self.rx_queues = [queue.Queue() for _ in range(len(self.usrp_config))]
        self.disp_queue = queue.Queue(maxsize=10000)
        
        ### Make Connections 
        self._connect_logging()
    
    def _init_ui(self): 
        ### Define main wndow
        self.setWindowTitle('BioView')
        scriptDir = os.path.dirname(os.path.realpath(__file__))
        self.setWindowIcon(QIcon(os.path.join(scriptDir, 'assets', 'icon.png')))
        screen = QGuiApplication.primaryScreen().geometry()
        width = screen.width()
        height = screen.height()
        self.setGeometry(int(0.2 * width), int(0.1 * height), int(0.6 * width), int(0.8 * height))
        
        ### Create central widget and main layout
        central_widget = QWidget(self)
        self.setCentralWidget(central_widget)
        main_layout = QVBoxLayout(central_widget)
        
        ### Top shelf container
        top_layout = QHBoxLayout()
        
        ### All controls are in one container    
        controls_layout = QVBoxLayout() 
    
        ### Connect/Start/Stop/Balance Signal Buttons 
        self.app_control_panel = AppControlPanel()
        controls_layout.addWidget(self.app_control_panel, stretch=1)
        # Connect signal handlers
        self.app_control_panel.connectionInitiated.connect(self.start_initialization)
        self.app_control_panel.startRequested.connect(self.start_recording)
        self.app_control_panel.stopRequested.connect(self.stop_recording)
        self.app_control_panel.saveRequested.connect(self.update_save_state)
        self.app_control_panel.instructionsEnabled.connect(self.toggle_instructions)
        self.app_control_panel.balanceRequested.connect(self.balance_signals)
        
        experiment_layout = QHBoxLayout()
        
        ### Experiment Control Panel
        self.experiment_settings_panel = ExperimentSettingsPanel(self.exp_config)
        experiment_layout.addWidget(self.experiment_settings_panel, stretch=1)
        # Connect handlers
        self.experiment_settings_panel.timeWindowChanged.connect(self.handle_time_window_change)
        self.experiment_settings_panel.gridLayoutChanged.connect(self.handle_grid_layout_change)
        self.experiment_settings_panel.addChannelRequested.connect(self.handle_add_channel)
        self.experiment_settings_panel.removeChannelRequested.connect(self.handle_remove_channel)
        
        ### USRP Device Config Panel(s)
        self.usrp_config_panel = [None] * len(self.usrp_config)
        for idx, usrp_cfg in enumerate(self.usrp_config): 
            self.usrp_config_panel[idx] = UsrpDeviceConfigPanel(usrp_cfg)
            experiment_layout.addWidget(self.usrp_config_panel[idx], stretch=1)
            
        controls_layout.addLayout(experiment_layout, stretch=4)
        top_layout.addLayout(controls_layout, stretch=3)   
        
        ### Metadata Panels
        self.meta_panels = QVBoxLayout() 
        # Status Panel - Experiment Log goes here
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        self.log_display_panel = LogDisplayPanel(logger=self.logger)
        self.meta_panels.addWidget(self.log_display_panel, stretch=3)
        
        # Annotation Panel
        self.annotate_event_panel = AnnotateEventPanel(self.exp_config)
        self.meta_panels.addWidget(self.annotate_event_panel, stretch=2)
        top_layout.addLayout(self.meta_panels, stretch=2)
        
        main_layout.addLayout(top_layout)
        
        ### Plot Grid - Initialized using config['disp_channels']
        self.plot_grid = PlotGrid(self.exp_config)
        main_layout.addWidget(self.plot_grid)
        
        # Init plot grid
        self.experiment_settings_panel._add_channels_to_grid()
        central_widget.setLayout(main_layout)
        
        ### Status Bar 
        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)
        self.device_status_panel = DeviceStatusPanel(self.devices)
        # Add device status panel to status bar (on the right side)
        self.status_bar.addPermanentWidget(self.device_status_panel)
        # Add some info text to status bar
        self.status_bar.showMessage("Ready")
          
    def _generate_usrp_mappings(self): 
        if self.usrp_config is None: 
            return
        
        ### [1] -  Generate channel internal:external mapping
        counter = 0 
        for idx, cfg in enumerate(self.usrp_config):
            self.usrp_config[idx].absolute_channel_nums = [counter + val for val in cfg.rx_channels]
            counter += len(cfg.rx_channels)
        
        ### [2] - Generate usrp channel pair labels 
        if getattr(self.exp_config, 'channel_mapping', None) is None: 
            num_usrp_devices = len(self.usrp_config)
            rx_per_usrp = [len(x.rx_channels) for x in self.usrp_config]
            tx_per_usrp = [len(x.tx_channels) for x in self.usrp_config]
            
            self.exp_config.channel_mapping = get_channel_map(
                n_devices = num_usrp_devices, 
                rx_per_dev = rx_per_usrp, 
                tx_per_dev = tx_per_usrp, 
                balance = getattr(self.exp_config, 'balance', False), 
                multi_pairs = getattr(self.exp_config, 'multi_pairs', None) 
            )
        
        ### [3] - Generate usrp channel pair labels:data queue index mapping
        ch_map = self.exp_config.channel_mapping
        for ridx in range(len(ch_map)):
            for tidx in range(len(ch_map[ridx])):
                label = ch_map[ridx][tidx]
                if label != '': 
                    self.exp_config.data_mapping[label] = self.source_counter
                    self.source_counter += 1
        
        ### [4] Add usrp Parameters
        # Store sampling rate
        self.exp_config.samp_rate = self.usrp_config[0].samp_rate
        # Populate list of all channel frequencies
        channel_ifs = [None] * len(self.exp_config.channel_mapping)
        for cfg in self.usrp_config:
            for idx, abs_idx in enumerate(cfg.absolute_channel_nums):  
                channel_ifs[abs_idx] = cfg.if_freq[idx]
        self.exp_config.channel_ifs = channel_ifs
        
    def _generate_biopac_mappings(self):
        if self.bio_config is None:
            return 
        
        ### Generate Biopac channel labels:data queue index mapping
        for ch in self.bio_config.channels: 
            label = f'{self.bio_config.device_name}_Ch{ch}'
            self.exp_config.data_mapping[label] = self.source_counter
            self.source_counter += 1
             
    def _connect_logging(self): 
        self.plot_grid.logEvent.connect(self.log_display_panel.log_message)
        for idx, panel in enumerate(self.usrp_config_panel): 
            panel.logEvent.connect(self.log_display_panel.log_message)
        
    def start_initialization(self):
        # Disable button during initialization
        self.running_status = RunningStatus.NOINIT
        self.connection_status = ConnectionStatus.CONNECTING
        self.update_buttons() 
         
        # USRP   
        for idx, config in enumerate(self.usrp_config):
            self.usrp_init_thread[idx] = UsrpController(config)  
            self.usrp_init_thread[idx].initSucceeded.connect(
                lambda usrp, tx_streamer, rx_streamer, idx=idx: self.on_usrp_init_success(
                    usrp=usrp, 
                    tx_streamer=tx_streamer, 
                    rx_streamer=rx_streamer, 
                    idx=idx
                )
            )
            self.usrp_init_thread[idx].initFailed.connect(self.on_init_failure)
            self.usrp_init_thread[idx].logEvent.connect(self.log_display_panel.log_message)
            
            dev_name = self.usrp_config[idx].device_name
            self.device_status_panel.update_device_state(device_name=dev_name, new_state=ConnectionStatus.CONNECTING)
            self.usrp_init_thread[idx].start()
            self.usrp_init_thread[idx].wait()

        # BIOPAC
        if self.bio_config is not None: 
            self.bio_init_thread = BiopacController(self.bio_config)
            self.bio_init_thread.initSucceeded.connect(self.on_bio_init_success)
            self.bio_init_thread.initFailed.connect(self.on_init_failure)
            self.bio_init_thread.logEvent.connect(self.log_display_panel.log_message)
            self.device_status_panel.update_device_state(device_name='BIOPAC', new_state=ConnectionStatus.CONNECTING)
            self.bio_init_thread.start()
            self.bio_init_thread.wait()
        
    def start_recording(self):
        # Update state 
        self.running_status = RunningStatus.RUNNING
        self.connection_status = ConnectionStatus.CONNECTED
        
        # Start streaming threads
        for idx, cfg in enumerate(self.usrp_config):
            self.usrp_tx_thread[idx] = UsrpTransmitter(
                config = cfg, 
                usrp = self.usrps[idx], 
                tx_streamer = self.tx_streamers[idx]
            )
            self.usrp_tx_thread[idx].logEvent.connect(self.log_display_panel.log_message)
            self.usrp_tx_thread[idx].start()
         
        # Start receiving threads
        for idx, cfg in enumerate(self.usrp_config):
            self.usrp_rx_thread[idx] = UsrpReceiver(
                usrp = self.usrps[idx], 
                config = cfg, 
                rx_streamer = self.rx_streamers[idx], 
                rx_queue = self.rx_queues[idx],
                running = self.running_status.value
            )
            self.usrp_rx_thread[idx].logEvent.connect(self.log_display_panel.log_message)
            self.usrp_rx_thread[idx].start()
        
        # Start display thread 
        self.display_thread = DisplayWorker(
            config = self.exp_config, 
            disp_queue = self.disp_queue, 
            running = self.running_status.value
        )
        self.display_thread.logEvent.connect(self.log_display_panel.log_message)
        self.display_thread.dataReady.connect(self.plot_grid.add_new_data)        
        self.display_thread.start()
        
        # Start saving thread 
        self.save_thread = SaveWorker(
            exp_config = self.exp_config, 
            usrp_config = self.usrp_config,
            rx_queues = self.rx_queues, 
            disp_queue = self.disp_queue, 
            running = self.running_status.value, 
            saving = self.saving_status
        )
        self.save_thread.logEvent.connect(self.log_display_panel.log_message)
        self.save_thread.start()
        
        # Start instructions
        if self.enable_instructions:
            self.instructions_thread = InstructionsWorker(config = self.exp_config)
            self.instructions_thread.textUpdate.connect(self.instruction_dialog.update_instruction_text)
            self.instructions_thread.logEvent.connect(self.log_display_panel.log_message)
            self.instructions_thread.start()
            
        # Start all threads together 
        
        
        # Update UI 
        self.update_buttons()
         
    def stop_recording(self):
        # Update state
        self.running_status = RunningStatus.STOPPED
        self.connection_status = ConnectionStatus.CONNECTED
        
        # Stop receiving threads
        for rx_thread in self.usrp_rx_thread:
            if rx_thread is not None: 
                rx_thread.stop()
        
        # Stop instruction 
        if self.instructions_thread is not None:
            self.instructions_thread.stop()
            
        # Stop saving thread
        if self.save_thread is not None: 
            self.save_thread.stop()
        
        # Stop display thread
        if self.display_thread is not None: 
            self.display_thread.stop()
        
        # Stop streaming threads
        for tx_thread in self.usrp_tx_thread:
            if tx_thread is not None: 
                tx_thread.stop()
    
        # Update UI 
        self.update_buttons()
     
    def balance_signals(self):
        # TODO: Implement signal balancing here 
        pass 
      
    def update_buttons(self): 
        self.app_control_panel.update_button_states(self.connection_status, self.running_status)
        self.experiment_settings_panel.update_button_states(self.connection_status, self.running_status)
    
    def on_usrp_init_success(self, usrp, tx_streamer, rx_streamer, idx=0):
        self.usrps[idx] = usrp
        self.tx_streamers[idx] = tx_streamer
        self.rx_streamers[idx] = rx_streamer
        
        # Update status bar
        dev_name = self.usrp_config[idx].device_name
        self.devices[dev_name] = ConnectionStatus.CONNECTED
        self.device_status_panel.update_device_state(device_name=dev_name, new_state=ConnectionStatus.CONNECTED)
        
        # If all devices have inited, do an overall status update
        self.on_init_success()
        
    def on_bio_init_success(self, biopac):
        self.biopac = biopac
        dev_name = self.bio_config.device_name
        self.devices[dev_name] = ConnectionStatus.CONNECTED
        self.device_status_panel.update_device_state(device_name=dev_name, new_state=ConnectionStatus.CONNECTED)
        
        # If all devices have inited, do an overall status update
        self.on_init_success()
        
    def on_init_success(self): 
        self.mutex.lock()
        inited = True
        for dev_name, dev_state in self.devices.items(): 
            if dev_state != ConnectionStatus.CONNECTED: 
                inited = False 
                break  
        
        if inited:    
            self.running_status = RunningStatus.STOPPED
            self.connection_status = ConnectionStatus.CONNECTED
            self.update_buttons() 
            
        self.mutex.unlock()
        
    def on_init_failure(self, error_message):
        self.running_status = RunningStatus.NOINIT
        self.connection_status = ConnectionStatus.DISCONNECTED
        self.update_buttons() 

        self.log_display_panel.log_message('error', error_message)
        self.usrp = None
    
    def handle_time_window_change(self, seconds):
        self.plot_grid.set_display_time(seconds)
    
    def handle_grid_layout_change(self, rows, cols):
        self.plot_grid.update_grid(rows, cols)
    
    def handle_add_channel(self, channel):
        self.plot_grid.add_channel(channel)
    
    def handle_remove_channel(self, channel):
        self.plot_grid.remove_channel(channel)
    
    def update_save_state(self, flag): 
        self.saving_status = flag
        
    def toggle_instructions(self, flag):
        self.enable_instructions = flag
        if self.instruction_dialog is not None: 
            self.instruction_dialog.toggle_ui(self.enable_instructions)
        
    def closeEvent(self, a0):
        # Ensure all threads are closed
        self.stop_recording()
        return super().closeEvent(a0)
    


from .app_control import AppControlPanel
from .experiment_settings import ExperimentSettingsPanel
from .plot_grid import PlotGrid
from .usrp_device_config import UsrpDeviceConfigPanel
from .log_display import LogDisplayPanel
from .annotate_event import AnnotateEventPanel
from .device_status import DeviceStatusPanel
from .text_dialog import TextDialog


import queue
import numpy as np
from collections import deque
import pyqtgraph as pg

from PyQt6.QtWidgets import QWidget, QGridLayout
from PyQt6.QtCore import pyqtSignal, QTimer, QEvent
from PyQt6.QtGui import QGuiApplication

from bioview.utils import get_color_by_idx

class PlotManager(): 
    def __init__(self, 
                disp_freq: int, 
                color: str,                  
                max_buffer_size: int, 
                display_duration: float, 
                data_src: str = '',
                xlabel: str = 'Time (s)',
                ylabel: str = 'Amplitude'
    ):
        # UI widget 
        self.widget = pg.PlotWidget()
        self.widget.setAntialiasing(False)
        self.widget.getPlotItem().setDownsampling(auto=True, mode='subsample')
        self.widget.setBackground(None) 
        self.widget.showGrid(x=True, y=True)
        self.widget.setLabel('bottom', xlabel)
        self.widget.setLabel('left', ylabel)
                
        # Create pen and plot item ONCE - this is key for performance
        self.pen = pg.mkPen(color=color, width=1)
        self.plot_item = self.widget.plot([], [], pen=self.pen)
        
        # Plot specs
        self.max_buffer_size = max_buffer_size
        self.display_duration = display_duration
        self.disp_freq = disp_freq
        
        # Data handling 
        self.data_src = data_src
        self.data_queue = queue.Queue()
        
        # Initialize after setting up basic properties
        self._init_plot()
        
    def _init_plot(self): 
        # Calculate number of points for the display duration
        self.num_points = int(self.display_duration * self.disp_freq)
        
        # Initialize buffer with zeros - deque with fixed maxlen for sliding window
        self.buffer = deque([0.0] * self.num_points, maxlen=self.num_points)
        
        # Create time vector that will be reused
        self.time_vector = np.linspace(0, self.display_duration, self.num_points, endpoint=False)
        
        # Set initial data on the plot item (don't create new plot)
        self.plot_item.setData(self.time_vector, list(self.buffer))
        
    def update_data_source(self, data_src: str = ''): 
        self.widget.setTitle(data_src)
        self.data_src = data_src
        self._init_plot()
        
        # Flush queue 
        while not self.data_queue.empty():
            try:
                self.data_queue.get_nowait()
            except queue.Empty:
                break
    
    def add_data(self, data): 
        """Add new data to the queue with overflow protection"""
        if isinstance(data, (list, np.ndarray)):
            # If it's an array, add each point individually to maintain proper sliding
            for point in np.atleast_1d(data):
                self._add_single_point(float(point))
        else:
            self._add_single_point(float(data))
    
    def _add_single_point(self, point):
        """Add a single point with queue size management"""
        # Keep queue size reasonable - if it gets too big, we're falling behind
        max_queue_size = self.num_points // 4  # Allow max 25% of display buffer in queue
        
        if self.data_queue.qsize() > max_queue_size:
            # We're falling behind - drop old data to stay real-time
            try:
                # Drop some old points to make room
                drop_count = max_queue_size // 4
                for _ in range(drop_count):
                    self.data_queue.get_nowait()
            except queue.Empty:
                pass
        
        self.data_queue.put(point)
        
    def update_plot(self): 
        # Calculate how many points we should process to stay real-time
        queue_size = self.data_queue.qsize()
        
        if queue_size == 0:
            return
            
        # If queue is getting large, process more aggressively to catch up
        if queue_size > self.num_points // 8:  # More than 12.5% of display buffer
            # Process more points to catch up, but skip some to stay real-time
            points_to_process = min(queue_size, self.num_points // 4)
            skip_ratio = max(1, queue_size // points_to_process)  # Skip every Nth point
        else:
            # Normal processing
            points_to_process = min(queue_size, 50)  # Process reasonable chunk
            skip_ratio = 1  # Don't skip any points
        
        updates_made = False
        points_processed = 0
        skip_counter = 0
        
        while not self.data_queue.empty() and points_processed < points_to_process:
            try:
                new_point = self.data_queue.get_nowait()
                
                # Skip points if we're behind (temporal downsampling)
                skip_counter += 1
                if skip_counter >= skip_ratio:
                    skip_counter = 0
                    # Add to deque - this automatically removes oldest point due to maxlen
                    self.buffer.append(new_point)
                    updates_made = True
                
                points_processed += 1
                
            except queue.Empty:
                break
        
        # Only update the plot if we have new data
        if updates_made:
            # Convert deque to list for plotting - this is the sliding window effect
            self.plot_item.setData(self.time_vector, list(self.buffer))

    def update_display_duration(self, duration): 
        self.display_duration = duration
        self._init_plot()

class PlotGrid(QWidget):
    logEvent = pyqtSignal(str, str)
    
    def __init__(self, 
                 config, 
                 parent = None):
        super().__init__(parent)
        self.config = config
        
        self.rows = 2
        self.cols = 2
        self.display_duration = 10.0 
        
        self.disp_freq = config.samp_rate / (config.save_ds * config.disp_ds)
        self.selected_channels = {} 
        
        # Set up the layout
        self.layout = QGridLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setSpacing(5)    
        
        # Keep track of available plots that are not connected to an output 
        self.available_slots = queue.Queue() 
        
        # Optimize refresh rate and ensure real-time performance
        self.refresh_time = max(self._get_monitor_refresh_delay(), 10)  # Max 60 FPS
        self.update_timer = QTimer()
        self.update_timer.timeout.connect(self.update_plots)
        
        # Add queue monitoring for debugging
        self.queue_monitor_timer = QTimer()
        self.queue_monitor_timer.timeout.connect(self._monitor_queues)
        self.queue_monitor_timer.start(2000)  # Check every 2 seconds
        
        # Initialize grid
        self.init_grid()
        self.update_timer.start(self.refresh_time)
        
    def _get_monitor_refresh_delay(self):
        screen = QGuiApplication.primaryScreen()
        if screen:
            return int(1000 // screen.refreshRate())
        else:
            return 16 # 60 Hz by default 

    # Handle theme changes     
    def event(self, event): 
        if event.type() == QEvent.Type.PaletteChange: 
            for r in range(self.rows):
                for c in range(self.cols): 
                    self.plots[r][c].widget.setBackground(None) 
                    self.plots[r][c].pen = pg.mkPen(color=get_color_by_idx(r*self.cols + c), width=1)
        return super().event(event)
    
    def init_grid(self):
        self.plots = [[None for _ in range(self.cols)] for _ in range(self.rows)] 
        
        for r in range(self.rows):
            for c in range(self.cols):
                plot_obj = PlotManager(
                    disp_freq=self.disp_freq,
                    color = get_color_by_idx(r*self.cols + c),
                    max_buffer_size = max(10, self.refresh_time // 10),  # Reasonable buffer size
                    display_duration = self.display_duration        
                )
                
                self.layout.addWidget(plot_obj.widget, r, c)
                self.plots[r][c] = plot_obj
                
                # Initially, all slots are available 
                self.available_slots.put((r, c))
    
    def update_grid(self, rows, cols):
        # Clear past grid 
        while(self.layout.count()): 
            child = self.layout.takeAt(0)
            if child.widget(): 
                child.widget().deleteLater()
        
        # Re-initialize 
        self.rows = rows 
        self.cols = cols 
        self.selected_channels = {} 
        
        # Flush queue of available slots 
        while not self.available_slots.empty():
            try:
                self.available_slots.get_nowait()
            except queue.Empty:
                break
        
        self.init_grid()
    
    def add_channel(self, channel):
        if channel in self.selected_channels.keys():
            self.logEvent.emit('debug', 'Unable to add channel as it is already being plotted')
            return
    
        try: 
            row, col = self.available_slots.get_nowait()
        except queue.Empty:
            self.logEvent.emit('warning', 'All graph slots full. Update layout or remove an existing trace.')
            return  
        
        plot_obj = self.plots[row][col]
        plot_obj.update_data_source(channel)
        
        # Update selected channel mapping
        self.selected_channels[channel] = {
            'plot': plot_obj,
            'loc': (row, col)
        }
    
    def remove_channel(self, channel):
        if channel not in self.selected_channels.keys():
            self.logEvent.emit('debug', 'Unable to remove channel as it is not being plotted')
            return
            
        # Clear the plot
        plot_obj = self.selected_channels[channel]['plot']
        loc = self.selected_channels[channel]['loc']
        
        plot_obj.update_data_source()
        
        # Remove from data structures
        self.selected_channels.pop(channel, None)
        self.available_slots.put(tuple(loc))
    
    def add_new_data(self, data): 
        """Add new data to the appropriate channels"""
        for channel_idx in range(np.shape(data)[0]): 
            if channel_idx >= len(self.config.disp_channels):
                break
                
            channel_key = self.config.disp_channels[channel_idx]
            if channel_key not in self.selected_channels.keys(): 
                continue 
                
            plot_obj = self.selected_channels[channel_key]['plot']
            # Pass the data for this channel (could be multiple samples)
            plot_obj.add_data(data[channel_idx, :])

    def update_plots(self):
        """Update all active plots"""
        for val in self.selected_channels.values():
            plot_obj = val['plot']
            plot_obj.update_plot()
    
    def _monitor_queues(self):
        """Monitor queue sizes for debugging lag issues"""
        total_queued = 0
        max_queue = 0
        
        for channel_data in self.selected_channels.values():
            plot_obj = channel_data['plot']
            queue_size = plot_obj.data_queue.qsize()
            total_queued += queue_size
            max_queue = max(max_queue, queue_size)
        
        if max_queue > 100:  # Threshold for concern
            self.logEvent.emit('debug', f'Plot queues getting large - max: {max_queue}, total: {total_queued}. Consider reducing data rate or increasing processing.')
        elif total_queued > 0:
            self.logEvent.emit('debug', f'Active plot queues - max: {max_queue}, total: {total_queued}')
            
    def set_display_time(self, dur):
        self.display_duration = dur
        for r in range(self.rows):
            for c in range(self.cols):
                self.plots[r][c].update_display_duration(dur)


import uhd 
import queue
import numpy as np
from PyQt6.QtCore import QThread, pyqtSignal

from bioview.constants import INIT_DELAY, SETTLING_TIME, SAVE_BUFFER_SIZE
from bioview.types import UsrpConfiguration

class ReceiveWorker(QThread):
    logEvent = pyqtSignal(str, str)

    def __init__(self, 
                 usrp, 
                 config: UsrpConfiguration, 
                 rx_streamer, 
                 rx_queue, 
                 running:bool = True,
                 parent = None
        ):
        super().__init__(parent)
        # Modifiable params
        self.rx_gain = config.get_param_value('rx_gain').copy()
        
        self.config = config 
        
        # Device params
        self.usrp = usrp
        self.rx_streamer = rx_streamer
        self.rx_queue = rx_queue
        self.running = running

    def run(self):
        self.logEvent.emit('debug', 'Receiving Started')
        if self.usrp is None or self.rx_streamer is None:
            self.logEvent.emit('error', 'USRP or Rx streamer not initialized.')
            return

        rx_metadata = uhd.types.RXMetadata() 
        
        # Buffer for receiving samples
        num_channels = self.rx_streamer.get_num_channels()
        max_samps_per_packet = self.rx_streamer.get_max_num_samps()
        
        recv_buffer_size = max_samps_per_packet * SAVE_BUFFER_SIZE # Make receive buffer larger than max_samps_per_packet
        
        recv_buffer = np.empty((num_channels, recv_buffer_size), dtype=np.complex64)
        
        # Setup streaming using continuous saving mode by default
        stream_cmd = uhd.types.StreamCMD(uhd.types.StreamMode.start_cont)
        
        # When using multiple devices, we need to set stream_now to False to align time edges of packets  
        stream_cmd.stream_now = False
        stream_cmd.time_spec = uhd.types.TimeSpec(self.usrp.get_time_now().get_real_secs() + INIT_DELAY)
        self.rx_streamer.issue_stream_cmd(stream_cmd)
        
        # Initialize 
        total_samps_received = 0 
        timeout = 0.5 # Larger timeout initially
        had_an_overflow = False
        last_overflow = uhd.types.TimeSpec(0)
    
        # Setup the statistic counters
        num_rx_samps = 0
        num_rx_dropped = 0
        
        rate = self.usrp.get_rx_rate()

        while self.running:
            # Check for updated parameters 
            curr_rx_gain = self.config.get_param_value('rx_gain')
            if curr_rx_gain != self.rx_gain: 
                for chan in self.config.rx_channels:
                    self.usrp.set_rx_gain(curr_rx_gain[chan], chan)
                self.logEvent.emit('debug', f'Rx gain updated to {curr_rx_gain}. Current {self.rx_gain}')
                self.rx_gain = curr_rx_gain
            
            try:
                # Receive samples
                num_rx_samps = self.rx_streamer.recv(recv_buffer, rx_metadata, timeout)
            except RuntimeError as ex:
                self.logEvent.emit('error', f'Receiver Runtime Eror: {ex}')
                continue

            timeout = INIT_DELAY # Reduce timeout for subsequent transmissions
        
            # Reference: uhd/examples/python/benchmark_rate.py
            # Handle the error codes
            if rx_metadata.error_code == uhd.types.RXMetadataErrorCode.none:
                # Reset the overflow flag
                if had_an_overflow:
                    had_an_overflow = False
                    num_rx_dropped += (rx_metadata.time_spec - last_overflow).to_ticks(rate)
            elif rx_metadata.error_code == uhd.types.RXMetadataErrorCode.overflow:
                had_an_overflow = True
                # Need to make sure that last_overflow is a new TimeSpec object, not
                # a reference to metadata.time_spec, or it would not be useful
                # further up.
                last_overflow = uhd.types.TimeSpec(
                    rx_metadata.time_spec.get_full_secs(), rx_metadata.time_spec.get_frac_secs()
                )
                self.logEvent.emit('warning', f'Receiver Overflow: {rx_metadata.strerror()}')
            elif rx_metadata.error_code == uhd.types.RXMetadataErrorCode.late:
                self.logEvent.emit('warning', f'Receiver Late: {rx_metadata.strerror()}, restarting...')
                # Radio core will be in the idle state. Issue stream command to restart streaming.
                stream_cmd.time_spec = uhd.types.TimeSpec(
                    self.usrp.get_time_now().get_real_secs() + INIT_DELAY
                )
                stream_cmd.stream_now = num_channels == 1
                self.rx_streamer.issue_stream_cmd(stream_cmd)
            elif rx_metadata.error_code == uhd.types.RXMetadataErrorCode.timeout:
                self.logEvent.emit('warning', f'Receiver Timeout: {rx_metadata.strerror()}')
            else:
                self.logEvent.emit('warning', f'Receiver Error: {rx_metadata.strerror()}')

            total_samps_received += num_rx_samps
            
            # Copy samples to avoid buffer overwrite and put in queue
            # recv_buffer.dtype = np.complex64 (since default cpu_format = 'fc32')
            try:
                self.rx_queue.put((recv_buffer))
            except queue.Full:
                self.logEvent.emit('warning', 'Rx Queue full, dropping buffer')
            except queue.Empty: 
                self.logEvent.emit('debug', 'Rx Queue Empty')
                continue
                
        # Gracefully close once receiving is finished
        stream_cmd = uhd.types.StreamCMD(uhd.types.StreamMode.stop_cont)
        self.rx_streamer.issue_stream_cmd(stream_cmd)
        self.logEvent.emit('debug', 'Receiving Stopped')
        
    def stop(self):
        self.running = False